/* xrp.c
 *
 *   SAPF kernel module
 *
 *   header file for parsing and formatting support for
 *   XRP (extensible resolution protocol)
 *
 * Copyright (C) 2003 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 *
 */

#ifndef XRP_H
#define XRP_H


/*
 * XRP message format: (see LUNAR Internet Draft for more details)

4-Bytes header:
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|     version   |      ttl      |      flags    |    reserved   |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

plus multiple parameters objects of the form (inspired from CASP):
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|       length (bytes)          |     class     |  class-type   |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                      ... contents ...                         |
|                                                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

plus end-of-list object
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|              0                |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


Possible combinations:

     parclass                     classtype
     -------------------       ---------
  0  request series         0  SAPF sel
  1  reply addr             0  SAPF sel/eth
  2  target                 0  SAPF sel
                            1  IPv4
                            2  IPv6
  3  fwd addr               0  SAPF sel/eth
  4  backwd addr            0  SAPF sel/eth
  5  src addr               1  IPv4
  6  src id                 0  SAPF sel


LUNAR RREQ example (total 84 Bytes):

  4B  (version=0, ttl=3, flags=0, reserved=0)
 12B  (Len=12, parclass=request series, partype=sel)
 20B  (Len=20, parclass=reply addr, partype=sel/eth)
  4B  (Len=4, parclass=reqstyle, partype=sel/eth)
  8B  (Len=8, parclass=target, partype=IPv4)
 20B  (Len=20, parclass=backwd addr, partype=sel/eth)
  8B  (Len=8, parclass=src addr, partype=IPv4)
 12B  (Len=12, parclass=src id, partype=sel)
 (total: 84 Bytes)

LUNAR RREP example (total 36 Bytes):

  4B  (version=1, plen=2, flags=0, reserved=0)
 20B  (Len=20, parclass=fwd addr, partype=sel/eth)
 12B  (Len=12, parclass=target id, partype=sel)

*/

#define XRP_VERSION_REQUEST	0
#define XRP_VERSION_REPLY	1

#define XRP_CLASS_REQ_SERIES    0
#define XRP_CLASS_RES_STYLE     1  // resolution style (desired response's partype)
#define XRP_CLASS_REPLY_ADDR    2  // where to send back the reply
#define XRP_CLASS_TARGET        3  // the thing to resolve
#define XRP_CLASS_ERROR         4  // reply error code
#define XRP_CLASS_FWD_ADDR      5  // reply value for CO queries
#define XRP_CLASS_BACKWD_ADDR   6  // opt. request param for CO queries
#define XRP_CLASS_SRC_ADDR      7  // opt. request param
#define XRP_CLASS_SRC_ID        8  // opt. request/reply param
#define XRP_CLASS_BW            9  // bandwidth requirement

#define XRP_PARTYPE_UNKNOWN       0
#define XRP_PARTYPE_SEL           1
#define XRP_PARTYPE_IPv4          2
#define XRP_PARTYPE_IPv6          3
#define XRP_PARTYPE_SEL_ETH       4
#define XRP_PARTYPE_SEL_UDP       5
#define XRP_PARTYPE_HID128        7
#define XRP_PARTYPE_COUNT		8 // update len_array in xrp.c if changed
#define XRP_PARTYPE_BW            6

struct xrp_head_s {
	uint8_t version;
	uint8_t ttl;
	uint8_t flags;
	//uint8_t reserved;
        //uint8_t seqn;             //sequence number, it seems useless
        uint8_t  hops;
        uint32_t src;             //source address
        uint32_t dst;             //destination address
};

// Parameters:

struct xrp_common_s {
	uint16_t len;
	uint8_t parclass;
	uint8_t partype;
};
struct xrp_bw_s {
        uint16_t len;
        uint8_t  parclass;
        uint8_t  partype;
        double   bw;
};

struct xrp_sel_s {
	uint16_t len;
	uint8_t parclass;
	uint8_t partype;
	sapfsel_t sel;
};

struct xrp_sel_eth_s {
	uint16_t len;
	uint8_t parclass;
	uint8_t partype;
	sapfsel_t sel;
	unsigned char eth[6];
//	unsigned char pad[2];
};

struct xrp_sel_udp_s {
	uint16_t len;
	uint8_t parclass;
	uint8_t partype;
	sapfsel_t sel;
	uint32_t ip;
	uint16_t udp;
//	uint16_t pad;
};

struct xrp_ipv4_s {
	uint16_t len;
	uint8_t parclass;
	uint8_t partype;
	uint32_t ip;
};


struct xrp_hid128_s {
	uint16_t len;
	uint8_t parclass;
	uint8_t partype;
	uint8_t hid[16];
};


typedef union {
	struct xrp_common_s param;
	struct xrp_sel_s sel;
	struct xrp_sel_eth_s seleth;
	struct xrp_sel_udp_s seludp;
	struct xrp_ipv4_s ipv4;
	struct xrp_hid128_s hid128;
        struct xrp_bw_s bandwidth;
} xrp_param_t;


#endif // XRP_H
