/* lunar_dhcp.c
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   DHCP interception and address self-config
 *
 * Copyright (C) 2003 <christian.tschudin@unibas.ch>
 *
 *   2003-11-20 initial version of dhcp intercept
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 */


static int asearch_rrep_handler(sapfpkt *p, void *dummy);
static void asearch_timeout(void *dummy);

struct bootphdr {
	unsigned char opcode;
	unsigned char hw;
	unsigned char hwlength;
	unsigned char hopcount;
	uint32_t xid;
	uint16_t secs;
	uint16_t flags;
	uint32_t ciaddr, yiaddr, server_ip, bootp_gw_ip;
	unsigned char hwaddr[16];
	unsigned char servername[64];
	unsigned char bootfile[128];
} ;

#define DHCP_DISCOVER 1
#define DHCP_OFFER    2
#define DHCP_REQUEST  3
#define DHCP_ACK      5
#define DHCP_NACK     6

static struct addr_search_s {
	sapfsel_t callback;		// local handler to remove
	struct netbox_timer_s timer;
	unsigned char attempts;		// decrementing, 0=inactive
	unsigned char verify;		// retries
	uint32_t ip;			// addr currently probed
	unsigned char done[32];		// bitset of visited addresses
	unsigned char bootp[sizeof(struct bootphdr)];
	char is_request;
} asearch;

// ----------------------------------------------------------------------

static void send_dhcp(char dhcptype, uint32_t addr)
{
	unsigned char buf[512];
	int len;
	struct iphdr *ih = (struct iphdr*) buf;
	struct udphdr *uh = (struct udphdr*) (buf + sizeof(struct iphdr));
	struct bootphdr *bh = (struct bootphdr*) (buf + sizeof(struct iphdr) + sizeof(struct udphdr));

        DEBUG(9, printk("delivering DHCP reply: %d %08x\n", dhcptype, ntohl(addr)));
	memset(buf, 0, sizeof(buf));
	ih->ihl = 5;
	ih->version = 4;
	ih->ttl = 2;
	ih->protocol = 17;
	ih->saddr = htonl(LUNAR_IP_SUBNET | 0xff);
	ih->daddr = 0xffffffff;
	uh->source = htons(67);
	uh->dest = htons(68);
	memcpy(bh, asearch.bootp, sizeof(asearch.bootp));

	bh->opcode = 2; // BOOTP reply
	bh->yiaddr = addr;
	bh->server_ip = htonl(LUNAR_IP_SUBNET | 0xff);

	len = (unsigned char*)(bh + 1) - buf;
	memcpy(buf+len, "\x63\x82\x53\x63", 4);
	len += 4;

	memcpy(buf+len, asearch.is_request ? "\065\001" : "\065\001", 2);
	len += 2; buf[len++] = dhcptype;

	memcpy(buf + len, "\001\004\xff\xff\xff\x00", 6); // mask
	len += 6;

	if (asearch.is_request) {
		memcpy(buf + len, "\063\004\x00\x00\x20\x00", 6); // lease
		len += 6;
/*
	memcpy(buf + len, "\003\004\x01\x02\x03\x06", 6); // dgw
	len += 6;
*/
	}

	memcpy(buf + len, "\xff", 1); // eolist
	len++;

	ih->tot_len = htons((__u16)len);
	uh->len = htons(len - sizeof(struct iphdr));

	ih->check = ip_fast_csum(buf, ih->ihl);

	netbox_l3deliver(lunardev->ifindex, buf, len);
}

static void asearch_send_probe(uint32_t addr)
{
	sapfsel_t series;

        DEBUG(9, printk("asearch_send_probe %08x\n", ntohl(addr)));

	asearch.ip = addr;
	asearch.callback = mkHandlerPointer(
			(sapfcallback_t) asearch_rrep_handler, 0);

	series = sapf_getUnusedSelector(mysapfhandle, SELCLASS_LOCAL);
	sapf_addREDIR(mysapfhandle, series, 0);
	send_rreq(conf_max_ring, series, asearch.ip,
		  asearch.callback, 0, 0, 0, 0, 0, 0);
	netbox_newtimer(&asearch.timer, asearch_timeout, 0, LUNAR_IP_PROBE_RETRY);
}


static int asearch_rrep_handler(sapfpkt *p, void *dummy)
{
	struct xrp_head_s *x = (struct xrp_head_s*) p->data;

        DEBUG(9, printk("asearch_rrep_handler (%d bytes)\n", p->len));

	if (p->len < sizeof(struct xrp_head_s) ||
					    x->version != XRP_VERSION_REPLY)
		return 0;

	sapf_deleteEntry(mysapfhandle, asearch.callback);
	netbox_deltimer(&asearch.timer);

	if (asearch.is_request) {
		asearch.attempts = 0;
		send_dhcp(DHCP_NACK, asearch.ip);
	} else {
		set_bit(ntohl(asearch.ip) & 0xff, asearch.done);
		asearch.attempts--;
		if (asearch.attempts > 0)
			asearch_send_probe(1 + ntohl(asearch.ip));
		// else we give up - we send no reply to the client's DISCOVER
	}

	return 0;
}


static void asearch_timeout(void *dummy)
{
        DEBUG(9, printk("asearch_timeout (attempts=%d, verify=%d)\n",
			asearch.attempts, asearch.verify));

	sapf_deleteEntry(mysapfhandle, asearch.callback);

	asearch.verify++;
	if (asearch.verify > LUNAR_IP_VERIFY_MAX) { // address seems to be clean
		asearch.attempts = 0;
		if (asearch.is_request) {
			// ... would have to also collect gateways
			send_dhcp(DHCP_ACK, asearch.ip);
		} else
			send_dhcp(DHCP_OFFER, asearch.ip);
	} else
		asearch_send_probe(asearch.ip);
}


static void lunar_do_dhcp(sapfpkt *pkt)
{
	unsigned char *cp;
	struct iphdr *ih = (struct iphdr*)pkt->data;
	struct udphdr *uh;
	struct bootphdr *bh;
	uint32_t rq = 0;

	DEBUG(10, printk("lunar DHCP\n"));

	uh = (struct udphdr*) (ih + 1); // XXX should look for IP options
	bh = (struct bootphdr*) ( uh + 1);
/*
	if (pkt->len <= (unsigned char*) (bh + 1) - (unsigned char*)pkt->data) {
		DEBUG(10, printk("  too short\n"));
		return;
	}
*/

	// verify that it's a DHCP packet:
	cp = (unsigned char*) (bh + 1) + 4;
	while ((cp - (unsigned char*)pkt->data) < pkt->len && *cp != 0xff) {
		if (!*cp) { cp++; continue; }
		if (*cp == 53) { cp += 2; break; }
		cp += 2 + cp[1];
	}
	if (cp - (unsigned char*)pkt->data >= pkt->len || *cp == 0xff) {
		DEBUG(10, printk("  did not find DHCP option %d %d %d\n",
				 (char*)(bh+1) -(char*)pkt->data,
				 (char*)(uh+1) -(char*)pkt->data,
				 (char*)(ih+1) -(char*)pkt->data));
		return;
	}
	if ( *cp != DHCP_DISCOVER && *cp != DHCP_REQUEST) {
		DEBUG(10, printk("  option type was %d\n", *cp));
		return;
	}

	if (asearch.attempts) {
		netbox_deltimer(&asearch.timer);
		sapf_deleteEntry(mysapfhandle, asearch.callback);
		asearch.attempts = 0;
	}
	asearch.verify = 0;

	asearch.is_request = *cp == DHCP_REQUEST;

	memcpy(asearch.bootp, bh, sizeof(struct bootphdr));
	asearch.attempts = 10;

	memset(asearch.done, 0, sizeof(asearch.done));
	set_bit(0, asearch.done);
	set_bit(254, asearch.done);
	set_bit(255, asearch.done);

	cp = (unsigned char*) (bh + 1) + 4;
	while ((cp - (unsigned char*)pkt->data) < pkt->len && *cp != 0xff) {
		if (!*cp) { cp++; continue; }
		if (*cp == 50) { cp += 2; rq = *(uint32_t*)cp; break; }
		cp += 2 + cp[1];
	}

	if (rq) {
		DEBUG(9, printk(" DHCP requested: %08x\n", ntohl(rq)));
		if ((LUNAR_IP_NETMASK & ntohl(rq)) != LUNAR_IP_SUBNET) {
			if (asearch.is_request) {
				send_dhcp(DHCP_NACK, rq);
				return;
			}
			rq = 0;
		}
	}
	if (rq)
		asearch.ip = rq;
	else {
		unsigned char hid;
		get_random_bytes(&hid, 1);
		if (!hid) hid++; else if (hid >= 254) hid--;
		asearch.ip = htonl(LUNAR_IP_SUBNET | hid);
	}

	asearch_send_probe(asearch.ip);
}


// eof
